"""
DoENUT: data.modifiers
These classes provide ways to manipulate dataset (filtering, scaling, etc).
"""
from .ortho_scaler import OrthoScaler
from .column_selector import ColumnSelector
from .duplicate_remover import DuplicateRemover
from .duplicate_averager import DuplicateAverager
